/*
 *  Input routines for sort and shuffle programs.
 *  Read lines (or, #if DBSORT, db records) into array of pointers,
 *  reallocating as necessary.  Fall back and punt to a temporary file
 *  if necessary.
 *
 *  Copyright 1988-2003
 *  Steve Summit, scs@eskimo.com
 *  This code may be freely redistributed and used
 *  so long as the author's name and this notice remain intact,
 *  and any modifications are marked as such.
 *
 *  See http://www.eskimo.com/~scs/src for possible updates.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#ifdef DBSORT
#include <db.h>
#else
#include "agetline.h"
#endif

#include "defs.h"
#include "sortmain.h"
#include "tmpfile.h"

#ifndef DBSORT
char **array = NULL;
#else
dbnode *array = NULL;
#endif

int nalloc = 0;
int nlines = 0;

extern struct tmpfile *nextemp;

extern int usedtemps;

#ifdef DBSORT
extern dbdesc dbd;
#endif

#ifdef ESUITE
extern int maxchunk;
#endif

#ifdef VMLIMIT

extern char *lmalloc();
extern char *lrealloc();

#else

#define lmalloc malloc
#define lrealloc realloc

#endif

input(fd, name)
#ifndef DBSORT
FILE *fd;
#else
struct dbfd *fd;
#endif
char *name;
{
#ifndef DBSORT
char *p;
#else
dbnode p;
#endif
int r;

while(TRUE)
	{
#ifdef ESUITE
	if(maxchunk > 0 && nlines == maxchunk)
		anothertemp();
#endif

	if(nlines >= nalloc)
		{
		int newalloc;
		char *newarray;
#ifndef DBSORT
		int elsize = sizeof(char *);
#else
		int elsize = sizeof(dbnode);
#endif

		if(nlines == 0)
			newalloc = 50;
		else	newalloc = nlines / 2 * 3;

#ifndef SAFEREALLOC
		if(array == NULL)
			newarray = lmalloc(newalloc * elsize);
		else
#endif
			newarray = lrealloc((char *)array, newalloc * elsize);

		if(newarray == NULL && nlines > 0)
			{
			newalloc = nlines + 10;
			newarray = lrealloc((char *)array, newalloc * elsize);
			}

		if(newarray == NULL)
			{
			anothertemp();
			assert(nlines < nalloc);
			}
		else	{
			nalloc = newalloc;
#ifndef DBSORT
			array = (char **)newarray;
#else
			array = (dbnode *)newarray;
#endif
			}
		}

#ifndef DBSORT

	p = agetline2x(fd, &r, (int *)NULL);

	if(p == NULL)
		{
		if(r == AGL_EOF)
			break;
		else if(r == AGL_NOMEM)
			{
			anothertemp();
			continue;
			}
		else	{
			fprintf(stderr, "%s: read error\n", progname);
			exit(1);
			}
		}
#else

	p = db_fgetent(fd, dbd, (struct dbnode *)NULL, DB_CHECKRET, &r);

	if(p == NULL)
		{
		if(r == DB_EOF)
			break;
		else if(r == DB_NOMEM)
			{
			anothertemp();
			continue;
			}
		else	{
			fprintf(stderr, "%s: read error\n", progname);
			exit(1);
			}
		}
#endif

	nlines++;

	array[nlines - 1] = p;
	}
}

/*
 *  Sort current in-memory array, write to temporary file, then
 *  zero out in-memory array (freeing all memory, of course) in
 *  preparation for another chunk.
 */

anothertemp()
{
FILE *fd;
int i;
struct tmpfile *newtemp;

#ifdef VMLIMIT
lmalreset();
#endif

stashfree();

mktmpname(nextemp->tf_filename, 'a');

if((fd = fopen(nextemp->tf_filename, "w")) == NULL)
	{
	fprintf(stderr, "%s: can't open temporary file %s: ",
					progname, nextemp->tf_filename);
	perror("");
	exit(1);
	}

#ifdef DEBUG
fprintf(stderr, "using temporary file %s\n", nextemp->tf_filename);
#endif

nextemp->tf_flags |= TF_INUSE;

dosort();

output(fd, nextemp->tf_filename);
nextemp->tf_nents = nlines;

if(fclose(fd) == EOF)
	{
	fprintf(stderr, "%s: error closing (writing?) temporary file %s: ",
					progname, nextemp->tf_filename);
	perror("");
	exit(1);
	}

usedtemps = TRUE;

for(i = 0; i < nlines; i++)
	{
#ifndef DBSORT
	free(array[i]);
#else
	db_nfree(array[i]);
#endif
	}

nlines = 0;

newtemp = tfalloc();

nextemp->tf_next = newtemp;
nextemp = newtemp;

stashalloc();
}
