/*
 *  Shuffle an array of n items (each of size elwidth).
 *  Pretty much the inverse of qsort().
 *  Uses the classic in-place algorithm:
 *  Swap element 1 with an element randomly chosen from 1..n,
 *  then swap element 2 with an element randomly chosen from 2..n,
 *  etc.
 *
 *  Copyright 1998-2003
 *  Steve Summit, scs@eskimo.com
 *  This code may be freely redistributed and used
 *  so long as the author's name and this notice remain intact,
 *  and any modifications are marked as such.
 *
 *  See http://www.eskimo.com/~scs/src for possible updates.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "defs.h"
#include "sort.h"

extern char *progname;

shuffle(array, n, elwidth)
char *array;
int n;
int elwidth;
{
int i;
char *p1, *p2;
char tmpbuf[32];
char *tmp = tmpbuf;

if(elwidth > 32)
	{
	tmp = malloc(elwidth);
	if(tmp == NULL)
		{
		/* shouldn't be noisy */
 		fprintf(stderr, "%s: out of memory\n", progname);
		return;	/* retval? */
		}
	}

for(i = 0; i < n-1; i++)
	{
	int i2 = randrange(n-i);

	if(i2 != 0)
		{
		p1 = array + i * elwidth;
		p2 = array + (i + i2) * elwidth;
		memcpy(tmp, p1, elwidth);
		memcpy(p1, p2, elwidth);
		memcpy(p2, tmp, elwidth);
		}
	}

if(tmp != tmpbuf)
	free(tmp);

/* return value? */
}

int randrange(n)
int n;
{
return rand() / (RAND_MAX / n + 1);
}
